@@ -1,4 +1,4 @@
-// Copyright (C) 2012-2020 Internet Systems Consortium, Inc. ("ISC")
+// Copyright (C) 2012-2022 Internet Systems Consortium, Inc. ("ISC")
 //
 // This Source Code Form is subject to the terms of the Mozilla Public
 // License, v. 2.0. If a copy of the MPL was not distributed with this
@@ -9,6 +9,7 @@
 
 #include <asiolink/io_address.h>
 #include <asiolink/io_service.h>
+#include <cc/data.h>
 #include <database/db_exceptions.h>
 #include <dhcp/duid.h>
 #include <dhcp/option.h>
@@ -143,13 +144,13 @@ public:
     /// The query created will return statistics for a single subnet
     ///
     /// @param subnet_id id of the subnet for which stats are desired
-    /// @throw BadValue if sunbet_id given is 0.
+    /// @throw BadValue if subnet_id given is 0.
     LeaseStatsQuery(const SubnetID& subnet_id);
 
     /// @brief Constructor to query for the stats for a range of subnets
     ///
     /// The query created will return statistics for the inclusive range of
-    /// subnets described by first and last sunbet IDs.
+    /// subnets described by first and last subnet IDs.
     ///
     /// @param first_subnet_id first subnet in the range of subnets
     /// @param last_subnet_id last subnet in the range of subnets
@@ -311,20 +312,6 @@ public:
     /// @return lease collection
     virtual Lease4Collection getLease4(const ClientId& clientid) const = 0;
 
-    /// @brief Returns existing IPv4 lease for a given client identifier,
-    /// HW address and subnet identifier.
-    ///
-    /// @todo Consider whether this function is needed or not. In the basic
-    /// DHCPv4 server implementation it is not used by the allocation engine.
-    ///
-    /// @param client_id A client identifier.
-    /// @param hwaddr Hardware address.
-    /// @param subnet_id A subnet identifier.
-    ///
-    /// @return A pointer to the lease or NULL if the lease is not found.
-    virtual Lease4Ptr getLease4(const ClientId& client_id, const HWAddr& hwaddr,
-                                SubnetID subnet_id) const = 0;
-
     /// @brief Returns existing IPv4 lease for specified client-id
     ///
     /// There can be at most one lease for a given client-id in a single
@@ -711,6 +698,38 @@ public:
     /// @return number of leases removed.
     virtual size_t wipeLeases6(const SubnetID& subnet_id) = 0;
 
+    /// @brief Checks if the IPv4 lease limits set in the given user context are exceeded.
+    /// Abstract method.
+    ///
+    /// @param user_context all or part of the lease's user context which, for the intents and
+    /// purposes of lease limiting should have the following format
+    /// (not all nodes are mandatory and values are given only as examples):
+    /// { "ISC": { "limits": { "client-classes": [ { "name": "foo", "address-limit": 2 } ],
+    ///                        "subnet": { "id": 1, "address-limit": 2 } } } }
+    ///
+    /// @return a string describing a limit that is being exceeded, or an empty
+    /// string if no limits are exceeded
+    virtual std::string checkLimits4(isc::data::ConstElementPtr const& user_context) const = 0;
+
+    /// @brief Checks if the IPv6 lease limits set in the given user context are exceeded.
+    /// Abstract method.
+    ///
+    /// @param user_context all or part of the lease's user context which, for the intents and
+    /// purposes of lease limiting should have the following format
+    /// (not all nodes are mandatory and values are given only as examples):
+    /// { "ISC": { "limits": { "client-classes": [ { "name": "foo", "address-limit": 2, "prefix-limit": 1 } ],
+    ///                        "subnet": { "id": 1, "address-limit": 2, "prefix-limit": 1 } } } }
+    ///
+    /// @return a string describing a limit that is being exceeded, or an empty
+    /// string if no limits are exceeded
+    virtual std::string checkLimits6(isc::data::ConstElementPtr const& user_context) const = 0;
+
+    /// @brief Checks if JSON support is enabled in the database.
+    /// Abstract method.
+    ///
+    /// @return true if there is JSON support, false otherwise
+    virtual bool isJsonSupported() const = 0;
+
     /// @brief Return backend type
     ///
     /// Returns the type of the backend (e.g. "mysql", "memfile" etc.)
@@ -761,6 +780,42 @@ public:
     /// support transactions, this is a no-op.
     virtual void rollback() = 0;
 
+    /// @brief Sets IO service to be used by the Lease Manager.
+    ///
+    /// @param io_service IOService object, used for all ASIO operations.
+    static void setIOService(const isc::asiolink::IOServicePtr& io_service) {
+        io_service_ = io_service;
+    }
+
+    /// @brief Returns pointer to the IO service.
+    static isc::asiolink::IOServicePtr& getIOService() {
+        return (io_service_);
+    }
+
+    // -- The following are memfile only, but defined in the base LeaseMgr for convenience. --
+
+    /// @brief Returns the class lease count for a given class and lease type.
+    ///
+    /// @param client_class client class for which the count is desired
+    /// @param ltype type of lease for which the count is desired. Defaults to
+    /// Lease::TYPE_V4.
+    ///
+    /// @return number of leases
+    virtual size_t getClassLeaseCount(const ClientClass& client_class,
+                                      const Lease::Type& ltype = Lease::TYPE_V4) const = 0;
+
+    /// @brief Recount the leases per class for V4 leases.
+    virtual void recountClassLeases4() = 0;
+
+    /// @brief Recount the leases per class for V6 leases.
+    virtual void recountClassLeases6() = 0;
+
+    /// @brief Clears the class-lease count map.
+    virtual void clearClassLeaseCounts() = 0;
+
+private:
+    /// The IOService object, used for all ASIO operations.
+    static isc::asiolink::IOServicePtr io_service_;
 };
 
 }  // namespace dhcp
